/***********************************************************************************************************************
 *
 * blueBill Mobile - open source birdwatching
 * ==========================================
 *
 * Copyright (C) 2009, 2010 by Tidalwave s.a.s. (http://www.tidalwave.it)
 * http://bluebill.tidalwave.it/mobile/
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: MarshallerContextImpl.java,v 199b23ac9304 2010/06/24 16:19:15 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.semantic.io.json.impl;

import javax.annotation.Nonnull;
import javax.annotation.concurrent.NotThreadSafe;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.IdentityHashMap;
import java.util.List;
import java.io.IOException;
import org.json.me.JSONException;
import org.json.me.JSONWriter;
import it.tidalwave.util.As;
import it.tidalwave.util.AsException;
import it.tidalwave.util.Id;
import it.tidalwave.util.Identifiable;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.semantic.io.json.Triple;
import it.tidalwave.semantic.io.json.spi.TripleMarshaller;
import it.tidalwave.semantic.io.json.spi.TripleMarshallerFactory;
import it.tidalwave.semantic.io.json.spi.impl.AsTripleMarshaller;
import it.tidalwave.mobile.media.Media;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: $
 *
 **********************************************************************************************************************/
@NotThreadSafe
public class MarshallerContextImpl
  {
    private final Comparator<Triple> tripleComparator = new Comparator<Triple>()
      {
        public int compare (final @Nonnull Triple t1, final @Nonnull Triple t2)
          {
            final int d1 = idFor(t1.getSubject()).compareTo(idFor(t2.getSubject()));

            return (d1 != 0) ? d1 : t1.getPredicate().compareTo(t2.getPredicate());
          }
      };

    private List<As> pendingEntities = new ArrayList<As>();

    private IdentityHashMap<Object, Id> idMapByObject = new IdentityHashMap<Object, Id>();

    private List<Triple> triples = new ArrayList<Triple>();
    
    private int idSequence;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public MarshallerContextImpl (final @Nonnull As entity)
      throws JSONException, IOException
      {
        visitEntity(entity);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void marshalPendingEntities (final @Nonnull JSONWriter jw)
      throws IOException, JSONException
      {
        for (final As entity : pendingEntities)
          {
            triples.addAll(findTripleMarshallerFor(entity).marshal(entity));
          }

        marshalTriples(jw, triples);
      }

    /*******************************************************************************************************************
     *
     * Marshal some triples sharing the same subject.
     *
     ******************************************************************************************************************/
    private void marshalTriples (final @Nonnull JSONWriter jw, final @Nonnull List<Triple> triples)
      throws IOException, JSONException
      {
        if (!triples.isEmpty())
          {
            Object prevSubject = null;
            Id prevPredicate = null;
            final List<Object> objects = new ArrayList<Object>();

            for (final Triple triple : sortedBySubjectAndPredicate(triples))
              {
                final Object subject = triple.getSubject();

                if (prevSubject != subject)
                  {
                    if (prevSubject != null)
                      {
                        if (!objects.isEmpty())
                          {
                            jw.key(prevPredicate.stringValue());
                            marshalObjects(jw, objects);
                          }
                        
                        jw.endObject();
                      }

                    jw.key(idFor((As)subject).stringValue()).object();
                    prevSubject = subject;
                    objects.clear();
                    prevPredicate = triple.getPredicate();
                  }
            
                final Id predicate = triple.getPredicate();

                if (!predicate.equals(prevPredicate))
                  {
                    jw.key(prevPredicate.stringValue());
                    marshalObjects(jw, objects);
                    objects.clear();
                  }

                objects.add(triple.getObject());
                prevPredicate = predicate;
              }

            if (!objects.isEmpty())
              {
                jw.key(prevPredicate.stringValue());
                marshalObjects(jw, objects);
              }

            jw.endObject();
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private Id idFor (final @Nonnull As object)
      {
        if (object instanceof Media)
          {
            try
              {
                return ((Media)object).get(Media.ID); // FIXME: refactor as as(Identifiable).getId()
              }
            catch (NotFoundException e)
              {
                throw new RuntimeException(e);
              }
          }

        try
          {
            return object.as(Identifiable.class).getId();
          }
        catch (AsException e)
          {
            Id id = idMapByObject.get(object);

            if (id == null)
              {
                id = new Id("_:x" + (idSequence++));
                idMapByObject.put(object, id);
              }

            return id;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static TripleMarshaller findTripleMarshallerFor (final @Nonnull As object)
      {
        try 
          {
            return Locator.find(TripleMarshallerFactory.class).findTripleMarshallerFor(object);
          }
        catch (NotFoundException ex)
          {
            return new AsTripleMarshaller();
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void visitEntity (final @Nonnull As entity)
      throws JSONException, IOException
      {
        enqueueEntity(entity);

        for (final Triple triple : findTripleMarshallerFor(entity).marshal(entity))
          {
            final Object object = triple.getObject();

            if (object instanceof As)
              {
                if (!visited(object))
                  {
                    visitEntity((As)object);
                  }
              }

            final As subject = triple.getSubject();

            if (!visited(subject))
              {
                visitEntity(subject);
              }
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private boolean visited (final @Nonnull Object object)
      {
        // don't use pendingEntities.contains(object) because we don't want to call equals()
        for (final Object o : pendingEntities)
          {
            if (o == object)
              {
                return true;
              }
          }

        return false;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void enqueueEntity (final @Nonnull As object)
      {
        if (!pendingEntities.contains(object))
          {
            pendingEntities.add(object);
          }
      }

    /*******************************************************************************************************************
     *
     * Marshal the object part of a predicate.
     *
     ******************************************************************************************************************/
    private void marshalObjects (final @Nonnull JSONWriter jw, final @Nonnull Collection<Object> objects)
      throws IOException, JSONException
      {
        jw.array();

        for (final Object object : objects)
          {
            if (object instanceof As)
              {
                Literals.marshalReference(jw, idFor((As)object));
              }
            else
              {
                Literals.marshalLiteral(jw, object);
              }
          }

        jw.endArray();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private List<Triple> sortedBySubjectAndPredicate (final @Nonnull List<Triple> triples)
      {
        final List<Triple> result = new ArrayList<Triple>(triples);
        Collections.sort(result, tripleComparator);

        return result;
      }
  }
